@@ -51,10 +51,12 @@
 
 QT_BEGIN_NAMESPACE
 
+Q_LOGGING_CATEGORY(lcDimmer, "qt.quick.controls.popup.dimmer")
+
 /*!
     \qmltype Popup
     \inherits QtObject
-    \instantiates QQuickPopup
+//!     \instantiates QQuickPopup
     \inqmlmodule QtQuick.Controls
     \since 5.7
     \ingroup qtquickcontrols2-popups
@@ -313,7 +315,7 @@ bool QQuickPopupPrivate::acceptTouch(const QTouchEvent::TouchPoint &point)
     if (point.id() == touchId)
         return true;
 
-    if (touchId == -1 && point.state() != Qt::TouchPointReleased) {
+    if (touchId == -1 && point.state() != QEventPoint::Released) {
         touchId = point.id();
         return true;
     }
@@ -372,11 +374,11 @@ bool QQuickPopupPrivate::handleMouseEvent(QQuickItem *item, QMouseEvent *event)
 {
     switch (event->type()) {
     case QEvent::MouseButtonPress:
-        return handlePress(item, event->windowPos(), event->timestamp());
+        return handlePress(item, event->scenePosition(), event->timestamp());
     case QEvent::MouseMove:
-        return handleMove(item, event->windowPos(), event->timestamp());
+        return handleMove(item, event->scenePosition(), event->timestamp());
     case QEvent::MouseButtonRelease:
-        return handleRelease(item, event->windowPos(), event->timestamp());
+        return handleRelease(item, event->scenePosition(), event->timestamp());
     default:
         Q_UNREACHABLE();
         return false;
@@ -390,17 +392,17 @@ bool QQuickPopupPrivate::handleTouchEvent(QQuickItem *item, QTouchEvent *event)
     case QEvent::TouchBegin:
     case QEvent::TouchUpdate:
     case QEvent::TouchEnd:
-        for (const QTouchEvent::TouchPoint &point : event->touchPoints()) {
+        for (const QTouchEvent::TouchPoint &point : event->points()) {
             if (!acceptTouch(point))
-                return blockInput(item, point.pos());
+                return blockInput(item, point.position());
 
             switch (point.state()) {
-            case Qt::TouchPointPressed:
-                return handlePress(item, item->mapToScene(point.pos()), event->timestamp());
-            case Qt::TouchPointMoved:
-                return handleMove(item, item->mapToScene(point.pos()), event->timestamp());
-            case Qt::TouchPointReleased:
-                return handleRelease(item, item->mapToScene(point.pos()), event->timestamp());
+            case QEventPoint::Pressed:
+                return handlePress(item, item->mapToScene(point.position()), event->timestamp());
+            case QEventPoint::Updated:
+                return handleMove(item, item->mapToScene(point.position()), event->timestamp());
+            case QEventPoint::Released:
+                return handleRelease(item, item->mapToScene(point.position()), event->timestamp());
             default:
                 break;
             }
@@ -454,7 +456,8 @@ bool QQuickPopupPrivate::prepareExitTransition()
     if (transitionState != ExitTransition) {
         // The setFocus(false) call below removes any active focus before we're
         // able to check it in finalizeExitTransition.
-        hadActiveFocusBeforeExitTransition = popupItem->hasActiveFocus();
+        if (!hadActiveFocusBeforeExitTransition)
+            hadActiveFocusBeforeExitTransition = popupItem->hasActiveFocus();
         if (focus)
             popupItem->setFocus(false);
         transitionState = ExitTransition;
@@ -471,25 +474,35 @@ void QQuickPopupPrivate::finalizeEnterTransition()
     if (focus)
         popupItem->setFocus(true);
     transitionState = NoTransition;
+    getPositioner()->reposition();
     emit q->openedChanged();
-    emit q->opened();
+    opened();
 }
 
 void QQuickPopupPrivate::finalizeExitTransition()
 {
     Q_Q(QQuickPopup);
     getPositioner()->setParentItem(nullptr);
-    popupItem->setParentItem(nullptr);
-    popupItem->setVisible(false);
+    if (popupItem) {
+        popupItem->setParentItem(nullptr);
+        popupItem->setVisible(false);
+    }
     destroyOverlay();
 
     if (hadActiveFocusBeforeExitTransition && window) {
         // restore focus to the next popup in chain, or to the window content if there are no other popups open
-        QQuickPopup *popup = nullptr;
-        if (QQuickOverlay *overlay = QQuickOverlay::overlay(window))
-            popup = QQuickOverlayPrivate::get(overlay)->stackingOrderPopups().value(0);
-        if (popup && popup->hasFocus()) {
-            popup->forceActiveFocus();
+        QQuickPopup *nextFocusPopup = nullptr;
+        if (QQuickOverlay *overlay = QQuickOverlay::overlay(window)) {
+            const auto stackingOrderPopups = QQuickOverlayPrivate::get(overlay)->stackingOrderPopups();
+            for (auto popup : stackingOrderPopups) {
+                if (QQuickPopupPrivate::get(popup)->transitionState != ExitTransition) {
+                    nextFocusPopup = popup;
+                    break;
+                }
+            }
+        }
+        if (nextFocusPopup && nextFocusPopup->hasFocus()) {
+            nextFocusPopup->forceActiveFocus();
         } else {
             QQuickApplicationWindow *applicationWindow = qobject_cast<QQuickApplicationWindow*>(window);
             if (applicationWindow)
@@ -506,6 +519,12 @@ void QQuickPopupPrivate::finalizeExitTransition()
     emit q->closed();
 }
 
+void QQuickPopupPrivate::opened()
+{
+    Q_Q(QQuickPopup);
+    emit q->opened();
+}
+
 QMarginsF QQuickPopupPrivate::getMargins() const
 {
     Q_Q(const QQuickPopup);
@@ -566,14 +585,13 @@ void QQuickPopupPrivate::setBottomMargin(qreal value, bool reset)
 
 /*!
     \since QtQuick.Controls 2.5 (Qt 5.12)
-    \qmlpropertygroup QtQuick.Controls::Popup::anchors
     \qmlproperty Object QtQuick.Controls::Popup::anchors.centerIn
 
     Anchors provide a way to position an item by specifying its
     relationship with other items.
 
     A common use case is to center a popup within its parent. One way to do
-    this is with the \l {Item::}{x} and \l {Item::}{y} properties. Anchors offer
+    this is with the \l[QtQuick]{Item::}{x} and \l[QtQuick]{Item::}{y} properties. Anchors offer
     a more convenient approach:
 
     \qml
@@ -595,7 +613,8 @@ void QQuickPopupPrivate::setBottomMargin(qreal value, bool reset)
     \note Popups can only be centered within their immediate parent or
     the window overlay; trying to center in other items will produce a warning.
 
-    \sa {Popup Positioning}, {Item::anchors}
+    \sa {Popup Positioning}, {Item::}{anchors}, {Using Qt Quick Controls types
+        in property declarations}
 */
 QQuickPopupAnchors *QQuickPopupPrivate::getAnchors()
 {
@@ -634,7 +653,6 @@ void QQuickPopupPrivate::setWindow(QQuickWindow *newWindow)
 
         QQuickControlPrivate *p = QQuickControlPrivate::get(popupItem);
         p->resolveFont();
-        p->resolvePalette();
         if (QQuickApplicationWindow *appWindow = qobject_cast<QQuickApplicationWindow *>(newWindow))
             p->updateLocale(appWindow->locale(), false); // explicit=false
     }
@@ -657,6 +675,11 @@ void QQuickPopupPrivate::reposition()
     getPositioner()->reposition();
 }
 
+QPalette QQuickPopupPrivate::defaultPalette() const
+{
+    return QQuickTheme::palette(QQuickTheme::System);
+}
+
 static QQuickItem *createDimmer(QQmlComponent *component, QQuickPopup *popup, QQuickItem *parent)
 {
     QQuickItem *item = nullptr;
@@ -694,6 +717,8 @@ static QQuickItem *createDimmer(QQmlComponent *component, QQuickPopup *popup, QQ
         if (component)
             component->completeCreate();
     }
+    qCDebug(lcDimmer) << "finished creating dimmer from component" << component
+        << "for popup" << popup << "with parent" << parent << "- item is:" << item;
     return item;
 }
 
@@ -720,6 +745,7 @@ void QQuickPopupPrivate::createOverlay()
 void QQuickPopupPrivate::destroyOverlay()
 {
     if (dimmer) {
+        qCDebug(lcDimmer) << "destroying dimmer" << dimmer;
         dimmer->setParentItem(nullptr);
         dimmer->deleteLater();
         dimmer = nullptr;
@@ -818,6 +844,12 @@ QQuickPopup::~QQuickPopup()
     d->popupItem = nullptr;
     delete d->positioner;
     d->positioner = nullptr;
+
+    // If the popup is destroyed before the exit transition finishes,
+    // the necessary cleanup (removing modal dimmers that block mouse events,
+    // emitting closed signal, etc.) won't happen. That's why we do it manually here.
+    if (d->transitionState == QQuickPopupPrivate::ExitTransition && d->transitionManager.isRunning())
+        d->finalizeExitTransition();
 }
 
 /*!
@@ -1527,53 +1559,6 @@ void QQuickPopup::resetFont()
     d->popupItem->resetFont();
 }
 
-
-/*!
-    \since QtQuick.Controls 2.3 (Qt 5.10)
-    \qmlproperty palette QtQuick.Controls::Popup::palette
-
-    This property holds the palette currently set for the popup.
-
-    Popup propagates explicit palette properties to its children. If you change a specific
-    property on a popup's palette, that property propagates to all of the popup's children,
-    overriding any system defaults for that property.
-
-    \code
-    Popup {
-        palette.text: "red"
-
-        Column {
-            Label {
-                text: qsTr("This will use red color...")
-            }
-
-            Switch {
-                text: qsTr("... and so will this")
-            }
-        }
-    }
-    \endcode
-
-    \sa Control::palette, ApplicationWindow::palette, {qtquickcontrols2-palette}{palette QML Basic Type}
-*/
-QPalette QQuickPopup::palette() const
-{
-    Q_D(const QQuickPopup);
-    return d->popupItem->palette();
-}
-
-void QQuickPopup::setPalette(const QPalette &palette)
-{
-    Q_D(QQuickPopup);
-    d->popupItem->setPalette(palette);
-}
-
-void QQuickPopup::resetPalette()
-{
-    Q_D(QQuickPopup);
-    d->popupItem->resetPalette();
-}
-
 QQuickWindow *QQuickPopup::window() const
 {
     Q_D(const QQuickPopup);
@@ -1836,6 +1821,8 @@ void QQuickPopup::setModal(bool modal)
         d->toggleOverlay();
     emit modalChanged();
 
+    QQuickItemPrivate::get(d->popupItem)->isTabFence = modal;
+
     if (!d->hasDim) {
         setDim(modal);
         d->hasDim = false;
@@ -1972,8 +1959,8 @@ void QQuickPopup::setOpacity(qreal opacity)
     This property holds the scale factor of the popup. The default value is \c 1.0.
 
     A scale of less than \c 1.0 causes the popup to be rendered at a smaller size,
-    and a scale greater than \c 1.0 renders the popup at a larger size. A negative
-    scale causes the popup to be mirrored when rendered.
+    and a scale greater than \c 1.0 renders the popup at a larger size. Negative
+    scales are not supported.
 */
 qreal QQuickPopup::scale() const
 {
@@ -2360,6 +2347,35 @@ void QQuickPopup::resetBottomInset()
     d->popupItem->resetBottomInset();
 }
 
+/*!
+    \since QtQuick.Controls 2.3 (Qt 5.10)
+    \qmlproperty palette QtQuick.Controls::Popup::palette
+
+    This property holds the palette currently set for the popup.
+
+    Popup propagates explicit palette properties to its children. If you change a specific
+    property on a popup's palette, that property propagates to all of the popup's children,
+    overriding any system defaults for that property.
+
+    \code
+    Popup {
+        palette.text: "red"
+
+        Column {
+            Label {
+                text: qsTr("This will use red color...")
+            }
+
+            Switch {
+                text: qsTr("... and so will this")
+            }
+        }
+    }
+    \endcode
+
+    \sa Item::palette, Window::palette, ColorGroup, Palette
+*/
+
 bool QQuickPopup::filtersChildMouseEvents() const
 {
     Q_D(const QQuickPopup);
@@ -2373,7 +2389,7 @@ void QQuickPopup::setFiltersChildMouseEvents(bool filter)
 }
 
 /*!
-    \qmlmethod QtQuick.Controls::Popup::forceActiveFocus(reason = Qt.OtherFocusReason)
+    \qmlmethod QtQuick.Controls::Popup::forceActiveFocus(enumeration reason = Qt.OtherFocusReason)
 
     Forces active focus on the popup with the given \a reason.
 
@@ -2409,6 +2425,13 @@ void QQuickPopup::componentComplete()
 
     d->complete = true;
     d->popupItem->componentComplete();
+
+    if (isVisible()) {
+        if (d->closePolicy & QQuickPopup::CloseOnEscape)
+            d->popupItem->grabShortcut();
+        else
+            d->popupItem->ungrabShortcut();
+    }
 }
 
 bool QQuickPopup::isComponentComplete() const
@@ -2550,7 +2573,7 @@ void QQuickPopup::fontChange(const QFont &newFont, const QFont &oldFont)
     emit fontChanged();
 }
 
-void QQuickPopup::geometryChanged(const QRectF &newGeometry, const QRectF &oldGeometry)
+void QQuickPopup::geometryChange(const QRectF &newGeometry, const QRectF &oldGeometry)
 {
     Q_D(QQuickPopup);
     d->reposition();
@@ -2582,6 +2605,7 @@ void QQuickPopup::itemChange(QQuickItem::ItemChange change, const QQuickItem::It
             else
                 d->popupItem->ungrabShortcut();
         }
+        break;
     default:
         break;
     }
@@ -2628,13 +2652,6 @@ void QQuickPopup::paddingChange(const QMarginsF &newPadding, const QMarginsF &ol
     }
 }
 
-void QQuickPopup::paletteChange(const QPalette &newPalette, const QPalette &oldPalette)
-{
-    Q_UNUSED(newPalette);
-    Q_UNUSED(oldPalette);
-    emit paletteChanged();
-}
-
 void QQuickPopup::spacingChange(qreal newSpacing, qreal oldSpacing)
 {
     Q_UNUSED(newSpacing);
@@ -2659,11 +2676,6 @@ QFont QQuickPopup::defaultFont() const
     return QQuickTheme::font(QQuickTheme::System);
 }
 
-QPalette QQuickPopup::defaultPalette() const
-{
-    return QQuickTheme::palette(QQuickTheme::System);
-}
-
 #if QT_CONFIG(accessibility)
 QAccessible::Role QQuickPopup::accessibleRole() const
 {
@@ -2682,10 +2694,10 @@ QString QQuickPopup::accessibleName() const
     return d->popupItem->accessibleName();
 }
 
-void QQuickPopup::setAccessibleName(const QString &name)
+void QQuickPopup::maybeSetAccessibleName(const QString &name)
 {
     Q_D(QQuickPopup);
-    d->popupItem->setAccessibleName(name);
+    d->popupItem->maybeSetAccessibleName(name);
 }
 
 QVariant QQuickPopup::accessibleProperty(const char *propertyName)
